//
// Created by webpigeon on 12/12/2021.
//

#ifndef FGGL_MATH_EASING_H
#define FGGL_MATH_EASING_H

#include <fggl/math/types.hpp>
#include <functional>

namespace fggl::math {

	constexpr inline float lerp(float a, float b, float w) {
		return (b - a) * w + a;
	}

	constexpr inline float scale(float in, float inMin, float inMax, float outMin, float outMax) {
		return ((in - inMin) * (outMax - outMin)) / (inMax - inMin);
	}

	//
	// Functions
	//
	using transformF = std::function<float(float)>;

	inline float scaleFilter(float in, float inMin, float inMax, float outMin, float outMax, const transformF& filter) {
		float out = in - inMin;
		out /= (inMax - inMin);
		out = filter(out);
		out *= (outMax - outMin);
		return out + outMin;
	}

	inline float mix(const transformF& funcA, const transformF& funcB, float weightB, float t) {
		return ((1 - weightB) * funcA(t)) + (weightB * funcB(t));
	}

	inline float crossFade(transformF funcA, transformF funcB, float t) {
		return ((1 - t) * funcA(t)) + (t * funcB(t));
	}

	//
	// building blocks
	//
	inline float scale(transformF f, float t) {
		return t * f(t);
	}

	inline float reverseScale(transformF f, float t) {
		return (1 - t) * f(t);
	}

	constexpr inline float Arch2(float t) {
		return t * (1 - t);
	}

	constexpr inline float BounceClampBottom(float t) {
		return fabs(t);
	}

	constexpr inline float BounceClampTop(float t) {
		return 1.f - fabs(1.f - t);
	}

	//
	// Easing function library
	// see Math for Game Programmers: Fast and Funky 1D Nonlinear Transformations, GDC 2015
	//

	constexpr inline float SmoothStart2(float t) {
		return t * t;
	}

	constexpr inline float SmoothStart3(float t) {
		return t * t * t;
	}

	constexpr inline float SmoothStart4(float t) {
		return t * t * t * t;
	}

	constexpr inline float SmoothStart5(float t) {
		return t * t * t * t * t;
	}

	constexpr inline float SmoothStop2(float t) {
		const auto tFlip = 1 - t;
		return 1 - (tFlip * tFlip);
	}

	constexpr inline float SmoothStop3(float t) {
		const auto tFlip = 1 - t;
		return 1 - (tFlip * tFlip);
	}

	constexpr inline float SmoothStop4(float t) {
		const auto tFlip = 1 - t;
		return 1 - (tFlip * tFlip * tFlip * tFlip);
	}

	constexpr inline float SmoothStop5(float t) {
		const auto tFlip = 1 - t;
		return 1 - (tFlip * tFlip * tFlip * tFlip * tFlip);
	}

	//
	// Bezier curves
	//

	constexpr inline float NormalizedBezier3(float B, float C, float t) {
		const float s = 1.f - t;
		const float t2 = t * t;
		const float s2 = s * s;
		const float t3 = t2 * t;
		return (3.f * B * s2 * t) + (3.f * C * s * t2) + t3;
	}

}

#endif //FGGL_UTILS_HPP
